Public Structure SectorIndex
    Public Row As Integer
    Public Column As Integer

    Public Sub New(ByVal R As Integer, ByVal C As Integer)
        Row = R
        Column = C
    End Sub
End Structure

Public Class Map
    Private mobjLayers() As MapLayer
    Private mobjEngine As Spritz3D


    Public ReadOnly Property Engine() As Spritz3D
        Get
            Return mobjEngine
        End Get
    End Property


    Friend Sub New(ByVal TheEngine As Spritz3D, ByVal LayerCount As Integer, ByVal SectorSize As Size, ByVal Dimentions As Size)
        Dim idx As Integer

        If LayerCount < 1 Then
            Throw New ArgumentOutOfRangeException("LayerCount", LayerCount, "Layer count must be a value of 1 or heigher.")
        End If

        ReDim mobjLayers(LayerCount - 1)

        For idx = mobjLayers.GetLowerBound(0) To mobjLayers.GetUpperBound(0)
            mobjLayers(idx) = New MapLayer(TheEngine, SectorSize, Dimentions)
        Next
    End Sub

    Public ReadOnly Property LayerCount() As Integer
        Get
            Return mobjLayers.Length
        End Get
    End Property

    Public Function GetLayer(ByVal Index As Integer) As MapLayer
        Return mobjLayers(Index)
    End Function
End Class

Public Class MapLayer
    Private mobjSectors(,) As MapSector
    Private mintBackgroundSprite As Integer
    Private mobjEngine As Spritz3D


    Public ReadOnly Property Engine() As Spritz3D
        Get
            Return mobjEngine
        End Get
    End Property

    Friend Sub New(ByVal TheEngine As Spritz3D, ByVal SectorSize As Size, ByVal Dimentions As Size)
        Dim idx, idy As Integer
        Dim mobjPosition As Point

        If Dimentions.Width < 1 Then
            Throw New ArgumentOutOfRangeException("Dimentions.Width", Dimentions.Width, "Dimentions.Width must be a value of 1 or heigher.")
        End If
        If Dimentions.Height < 1 Then
            Throw New ArgumentOutOfRangeException("Dimentions.Height", Dimentions.Height, "Dimentions.Height must be a value of 1 or heigher.")
        End If

        mobjEngine = TheEngine
        ReDim mobjSectors(Dimentions.Width - 1, Dimentions.Height - 1)
        For idy = 0 To Dimentions.Height - 1
            For idx = 0 To Dimentions.Width - 1
                mobjPosition = New Point(idx * SectorSize.Width, idy * SectorSize.Height)
                mobjSectors(idx, idy) = New MapSector(TheEngine, New Rectangle(mobjPosition, SectorSize))
            Next
        Next
    End Sub

    Public Function GetSector(ByVal Index As SectorIndex) As MapSector
        Return mobjSectors(Index.Row, Index.Column)
    End Function

    Public Property BackgroundSprite() As Integer
        Get
            Return mintBackgroundSprite
        End Get
        Set(ByVal Value As Integer)
            mintBackgroundSprite = Value
        End Set
    End Property
End Class

Public Class MapSector
    Implements IEnumerable, IDisposable

    Private mobjBounds As Rectangle
    Private mobjSprites, mobjFirstSprite, mobjLastSprite As LinkedLists.SpriteLinkedIndexListItem
    Private mintSpriteCount As Integer
    Private mobjEngine As Spritz3D


    Public ReadOnly Property Engine() As Spritz3D
        Get
            Return mobjEngine
        End Get
    End Property

    Friend Sub New(ByVal TheEngine As Spritz3D, ByVal Bounds As Rectangle)
        mobjEngine = TheEngine
        mobjBounds = Bounds
    End Sub

    Public ReadOnly Property SpriteCount() As Integer
        Get
            Return mintSpriteCount
        End Get
    End Property

    Public Function GetSprite(ByVal WithIndex As Integer) As Interfaces.ISprite
        Dim Tmp As LinkedLists.SpriteLinkedIndexListItem

        If mobjFirstSprite Is Nothing Then Return Nothing

        Tmp = mobjFirstSprite
        Do
            If Tmp.Index = WithIndex Then
                Return CType(mobjEngine.GetObject(Tmp.Index), Interfaces.ISprite)
            Else
                Tmp = Tmp.NextSprite
            End If
        Loop Until (Tmp.NextSprite Is Nothing)

        Return Nothing
    End Function

    Public Sub AddSprite(ByVal Index As Integer)
        If mobjSprites Is Nothing Then
            mobjSprites = New LinkedLists.SpriteLinkedIndexListItem()
            mobjSprites.Index = Index
            mobjFirstSprite = mobjSprites
            mobjLastSprite = mobjSprites
            mintSpriteCount += 1
            Exit Sub
        End If

        mobjLastSprite.NextSprite = New LinkedLists.SpriteLinkedIndexListItem()
        mobjLastSprite.NextSprite.Index = Index
        mobjLastSprite = mobjLastSprite.NextSprite
        mintSpriteCount += 1
    End Sub

    Public Sub RemoveSprite(ByVal WithIndex As Integer)
        Dim Tmp, objNext, objPrev As LinkedLists.SpriteLinkedIndexListItem

        Tmp = mobjFirstSprite
        Do

            If Tmp.Index = WithIndex Then
                objNext = Tmp.NextSprite
                objPrev = Tmp.PreviousSprite
                If Not objPrev Is Nothing Then objPrev.NextSprite = Tmp.NextSprite
                If Not objNext Is Nothing Then objNext.PreviousSprite = Tmp.PreviousSprite
                If Tmp Is mobjLastSprite Then mobjLastSprite = Tmp.PreviousSprite
                If Tmp Is mobjFirstSprite Then mobjFirstSprite = Tmp.NextSprite

                Tmp = Nothing
                mintSpriteCount -= 1
                Exit Do
            Else
                Tmp = Tmp.NextSprite
            End If

        Loop Until Tmp.NextSprite Is Nothing
    End Sub

    Public Sub MoveSprite(ByVal WithIndex As Integer, ByVal FromSector As MapSector, ByVal ToSector As MapSector)
        FromSector.RemoveSprite(WithIndex)
        ToSector.AddSprite(WithIndex)
    End Sub

    Public ReadOnly Property Bounds() As Rectangle
        Get
            Return mobjBounds
        End Get
    End Property

    Public Function GetSprites() As System.Collections.IEnumerator Implements System.Collections.IEnumerable.GetEnumerator
        Return New SpriteEnumerator(Me, mobjFirstSprite)
    End Function

    Public Sub Dispose() Implements System.IDisposable.Dispose
        GC.SuppressFinalize(Me)
    End Sub

    Public Class SpriteEnumerator
        Implements IEnumerator

        Private CurrentEntry, mobjFirstSpriteGiven As LinkedLists.SpriteLinkedIndexListItem
        Private mobjParent As MapSector


        ' Create a reference back to the object instance that 
        ' owns this Enumberator.
        Friend Sub New(ByVal Parent As MapSector, ByVal Sprites As LinkedLists.SpriteLinkedIndexListItem)
            CurrentEntry = Nothing
            mobjParent = Parent
            mobjFirstSpriteGiven = Sprites
        End Sub

        Public ReadOnly Property Current() As Object Implements System.Collections.IEnumerator.Current
            Get
                Return mobjParent.Engine.GetObject(CurrentEntry.Index)
            End Get
        End Property

        Public Function MoveNext() As Boolean Implements System.Collections.IEnumerator.MoveNext
            ' If we've itterated until we've reached the last one,
            ' then stop!
            If CurrentEntry Is Nothing Then
                CurrentEntry = mobjFirstSpriteGiven
                If CurrentEntry Is Nothing Then Return False Else Return True
            End If

            If Not CurrentEntry.NextSprite Is Nothing Then
                CurrentEntry = CurrentEntry.NextSprite
                Return True
            Else
                Return False
            End If
        End Function

        Public Sub Reset() Implements System.Collections.IEnumerator.Reset
            CurrentEntry = Nothing
        End Sub
    End Class
End Class